#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys
import time
import shutil
import hashlib
import datetime
import configparser

import requests
from requests.auth import HTTPBasicAuth

# check if we can find the cacerts, if not, take our own
try:
    requests.get("https://notar.direct/")
except requests.exceptions.SSLError:
    os.environ['REQUESTS_CA_BUNDLE'] = "cacert.pem"


def out(txt):
    """
    minimal function for logging

    works like tee into a logfile "process.log" and stdout

    :param txt: text to log
    """
    with open("process.log", "a") as log:
        log.write(txt)
    print(txt, end="")
    sys.stdout.flush()


def timestamp():
    """
    current timestamp for logfiles

    :return: human readable timestamp
    """
    return datetime.datetime.now().isoformat()


def hash_of_file(filename, blocksize=65536):
    """
    :param filename:
    :param blocksize:
    :return: the hash of the file
    """
    hasher = hashlib.sha256()
    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(blocksize), b""):
            hasher.update(chunk)
    return hasher.hexdigest()


def handle_has_file(filename):
    """read hashes stored in .has file

    .has files look something like this:
    hash=deadbeef...
    HASH = 123sda...

    :param filename: name of .has file to read
    :return: all hashes from the .has file
    """
    hashes = []
    with open(filename) as f:
        for line in f:
            try:
                key, value = line.split("=")
            except ValueError:
                continue
            if key.lower() == "hash":
                hashes.append(value.strip())
    return hashes


def scan_directory(directory):
    """
    searches the directory for files and handles them

    :param directory: directory to look at
    :return: list of all hashes gained from the directory
    """
    hashes = []
    files_to_move = []

    for name in os.listdir(directory):
        path = os.path.join(directory, name)
        if os.path.isfile(path):
            files_to_move.append(path)
            if name[-4:].lower() == ".has":
                added_hashes = handle_has_file(path)
            else:
                added_hashes = [hash_of_file(path)]
            out("%s: %s -- %r\n" % (timestamp(), name, added_hashes))
            hashes += added_hashes

    return hashes, files_to_move


def upload_hashes(api_key, hashes):
    """
    uploads hashes to NOTAR.direct

    :param api_key: API key to access NOTAR.direct
    :param hashes: list of hashes to upload
    :return: http status code
    """

    try:
        r = requests.post("https://notar.direct/api/v1/hash",
                          auth=HTTPBasicAuth(api_key, ''),
                          json={"hash": hashes})
    except requests.exceptions.SSLError:  # if we do not have a cacert, do upload anyway
        r = requests.post("https://notar.direct/api/v1/hash",
                          auth=HTTPBasicAuth(api_key, ''),
                          json={"hash": hashes},
                          verify=False)

    for k, v in r.json().items():
        if v == "invalid Hash":
            out("Failed: %s\n" % k)

    return r.status_code


def force_move(srcs, dst):
    """
    Move the files away, override if necessary, delete if wanted (dst=="DELETE")

    :param srcs: list of source files
    :param dst: destination directory
    """
    if dst == "DELETE":
        for f in srcs:
            os.remove(f)
    else:
        os.makedirs(dst, exist_ok=True)
        for f in srcs:
            try:
                shutil.move(f, dst)
            except shutil.Error:
                shutil.copy2(f, dst)
                os.remove(f)


if __name__ == '__main__':
    ini_file = "notarDirectUploader.ini"
    config = configparser.ConfigParser()
    try:
        config.read(ini_file)
    except:
        out("error in %s\n" % ini_file)
        sys.exit()
    else:
        in_path = config["default"]["inPath"]
        in_path = os.path.expanduser(in_path)
        out_path = config["default"]["outPath"]
        out_path = os.path.expanduser(out_path)
        api_key = config["default"]["APIKey"]
        try:
            pause = float(config["default"]["pause"])
        except:
            pause = 60

    os.makedirs(os.path.join(out_path), exist_ok=True)

    out("Starting at %s\n" % timestamp())
    out("    %s -> %s\n" % (in_path, out_path))
    out("    API key: %s\n" % api_key)
    out("    Pause: %ds\n" % pause)

    while True:
        hashes, files_to_move = scan_directory(in_path)

        if hashes:
            status = 0
            while status != 201:
                status = upload_hashes(api_key, hashes)
                out("Status: %d\n" % status)
                if status == 401:
                    out("ERROR: Your API Key is invalid\n")
                    sys.exit(1)
                if 500 <= status < 600:  # Server Error, try again later
                    time.sleep(pause)
                if 400 <= status < 500:  # Client Error, (not a single valid Hash)
                    break  # give up
            force_move(files_to_move, out_path)

        time.sleep(pause)
